from __future__ import annotations

import re
from datetime import datetime, date
from importlib.resources import files
from pathlib import Path


def sanitise_id(value):
    return re.sub(r'\W|^(?=\d)', '_', value).lower()


def date_int_w_ordinal(n: int):
    """Convert an integer to its ordinal as a string, e.g. 1 -> 1st, 2 -> 2nd, etc."""
    return str(n) + ('th' if 4 <= n % 100 <= 20 else {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, 'th'))


def ordinal_dt(dt: datetime | date) -> str:
    """Convert a datetime or date to a string with an ordinal day, e.g. 'Mon 1st Jan 2020'."""
    return dt.strftime(f'%a {date_int_w_ordinal(dt.day)} %b %Y')


def get_ui() -> Path:
    res = Path(files('shipaw'))
    res = res / 'ui'
    if not res.exists():
        raise FileNotFoundError(f'UI directory {res} does not exist')
    return res


def make_nice_str(s: str) -> str:
    return s.replace('_', ' ').title()



from __future__ import annotations

import functools
import os
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import quote

import pydantic as _p
from fastapi.encoders import jsonable_encoder
from loguru import logger
from pydantic import Field, computed_field
from pydantic_settings import BaseSettings, SettingsConfigDict
from starlette.templating import Jinja2Templates

from shipaw.fapi.ui_funcs import get_ui, ordinal_dt, sanitise_id
from shipaw.models.address import Address, Contact, FullContact
from shipaw.utils.consts_enums import ShipDirection
from shipaw.providers.registry import PROVIDER_TYPE_REGISTER, register_provider_instance

SHIPAW_ENV_KEY = 'SHIPAW_ENV'


def path_from_env_key(env_key: str) -> Path:
    env = os.getenv(env_key)
    if not env:
        raise ValueError(f'{env_key} not set')
    env_path = Path(env)
    if not env_path.exists():
        raise FileNotFoundError(f'{env_key} file {env_path} does not exist')
    return env_path


@dataclass
class ProviderEnv:
    name: str
    env_path: Path


ProviderEnvs = list[ProviderEnv]


@functools.cache
def get_templates_cached(template_dir: Path):
    temps = Jinja2Templates(directory=str(template_dir))
    temps.env.filters['jsonable'] = jsonable_encoder
    temps.env.filters['urlencode'] = lambda value: quote(str(value))
    temps.env.filters['sanitise_id'] = sanitise_id
    temps.env.filters['ordinal_dt'] = ordinal_dt
    return temps


class ShipawSettings(BaseSettings):
    # toggles
    shipper_live: bool = True
    log_level: str = 'DEBUG'

    # dirs
    data_dir: Path = Path.home() / 'shipaw'
    label_dir: Path = Path.home() / 'shipaw' / 'labels'
    log_db_path: str | None = None
    ui_dir: Path = Field(default_factory=get_ui)

    # Provider env file dict (json string in .env)
    provider_env_dict: dict[str, Path]
    default_provider_name: str | None = None

    # auto dirs
    # templates: Jinja2Templates | None = None

    # sender details
    address_line1: str
    address_line2: str | None = None
    address_line3: str | None = None
    town: str
    postcode: str
    country: str = 'GB'
    business_name: str
    contact_name: str
    email: str
    phone: str | None = None
    mobile_phone: str

    model_config = SettingsConfigDict(frozen=True)

    @classmethod
    @functools.cache
    def from_env(cls) -> ShipawSettings:
        logger.info(f'Loading ShipawSettings from env key: {SHIPAW_ENV_KEY}')
        env_path = path_from_env_key(SHIPAW_ENV_KEY)
        return cls(_env_file=env_path)  # pycharm_pydantic false positive

    #
    # @model_validator(mode='after')
    # def populate_provider_registry(self):
    #     logger.debug('Populating provider registry with providers from env')
    #     for name, env_path in self.provider_env_dict.items():
    #         logger.debug(f'Loading provider {name} from env file {env_path}')
    #         if provider_type := PROVIDER_TYPE_REGISTER.get(name):
    #             logger.debug(f'Found provider type {provider_type} for name {name}')
    #             provider_settings = provider_type.settings_type(_env_file=env_path)
    #             provider = provider_type(shipaw_settings=provider_settings)
    #             register_provider_instance(provider)
    #     return self

    @property
    def log_dir(self):
        return self.data_dir / 'logs'

    @property
    def template_dir(self):
        return self.ui_dir / 'templates'

    @property
    def static_dir(self):
        return self.ui_dir / 'static'

    @property
    def templates(self):
        return get_templates_cached(self.template_dir)

    ## SET LOGGING & LABELS ##
    @computed_field
    @property
    def log_file(self) -> Path:
        return self.log_dir / 'shipaw.log'

    @computed_field
    @property
    def ndjson_log_file(self) -> Path:
        return self.log_dir / 'shipaw.ndjson'

    @_p.model_validator(mode='after')
    def create_log_files(self):
        self.log_dir.mkdir(parents=True, exist_ok=True)
        for v in (self.log_file, self.ndjson_log_file):
            v.touch()
        return self

    @_p.field_validator('label_dir', mode='after')
    def create_label_dirs(cls, v, values):
        # todo bad path crashes progqram - try/except + fallback label path?
        directions = [_ for _ in ShipDirection]
        try:
            make_label_dirs(directions, v)
        except FileNotFoundError:
            v = Path.home() / 'Shipping Labels'
            make_label_dirs(directions, v)
        return v

    # SET ADDRESS/CONTACT OBJECTS #
    @property
    def contact(self):
        return Contact(
            contact_name=self.contact_name,
            email_address=self.email,
            mobile_phone=self.mobile_phone,
        )

    @property
    def address(self):
        return Address(
            address_lines=[_ for _ in [self.address_line1, self.address_line2, self.address_line3] if _],
            town=self.town,
            postcode=self.postcode,
            country=self.country,
            business_name=self.business_name,
        )

    @property
    def full_contact(self) -> FullContact:
        return FullContact(
            address=self.address,
            contact=self.contact,
        )


def make_label_dirs(directions, v):
    for direction in directions:
        apath = v / direction
        if not apath.exists():
            apath.mkdir(parents=True, exist_ok=True)


SHIPAW_SETTINGS = ShipawSettings.from_env()


def populate_providers(settings: ShipawSettings = None):
    settings = settings or SHIPAW_SETTINGS
    for name, env_path in settings.provider_env_dict.items():
        if provider_type := PROVIDER_TYPE_REGISTER.get(name):
            provider_settings = provider_type.settings_type(_env_file=env_path)
            register_provider_instance(provider_type(settings=provider_settings))



from collections.abc import Callable

import sqlalchemy
from pydantic import BaseModel
from sqlmodel import Column, Field

JSONTypes = str | list[str] | dict[str, str] | tuple[str, ...] | None
JSONTypesPydantic = BaseModel | list[BaseModel] | dict[str | int, BaseModel] | tuple[BaseModel, ...] | None


class PydanticJSONColumn(sqlalchemy.TypeDecorator):
    impl = sqlalchemy.JSON

    def __init__(self, model_class: type[BaseModel], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_class = model_class

    def process_bind_param(self, value: JSONTypesPydantic, dialect) -> JSONTypes:
        if value is None:
            return None
        elif isinstance(value, list):
            return [item.model_dump_json(round_trip=True) if isinstance(item, BaseModel) else item for item in value]
        elif isinstance(value, dict):
            return {
                key: item.model_dump_json(round_trip=True) if isinstance(item, BaseModel) else item
                for key, item in value.items()
            }
        elif isinstance(value, tuple):
            return tuple(
                item.model_dump_json(round_trip=True) if isinstance(item, BaseModel) else item for item in value
            )
        elif isinstance(value, BaseModel):
            return value.model_dump_json(round_trip=True)
        return None
        # elif isinstance(value, date):
        #     logger.debug(f'Processing date {value}')
        #     return value.isoformat()

    def process_result_value(self, value: JSONTypes, dialect) -> JSONTypesPydantic:
        if value is None:
            return None
        elif isinstance(value, list):
            return [self.model_class.model_validate_json(item) for item in value]
        elif isinstance(value, dict):
            return {key: self.model_class.model_validate_json(item) for key, item in value.items()}
        elif isinstance(value, tuple):
            return tuple(self.model_class.model_validate_json(item) for item in value)
        return self.model_class.model_validate_json(value)


def pydantic_json_column(model_class: type[BaseModel]):
    return Column(PydanticJSONColumn(model_class))


def required_json_field(model_class: type[BaseModel]):
    return Field(..., sa_column=pydantic_json_column(model_class))


def optional_json_field(model_class: type[BaseModel]):
    return Field(None, sa_column=pydantic_json_column(model_class))


def default_json_field(model_class: type[BaseModel], default_factory: Callable[[], JSONTypesPydantic]):
    return Field(default_factory=default_factory, sa_column=pydantic_json_column(model_class))



from typing import Any, TYPE_CHECKING
from collections.abc import Callable

if TYPE_CHECKING:
    from shipaw.fapi.requests import ShipmentRequest
    from shipaw.fapi.responses import ShipmentResponse

CALLBACK_REGISTER: dict[str, Callable[['ShipmentRequest', 'ShipmentResponse'], Any]] = {}



from pprint import pformat

from loguru import logger
from royal_mail_combined.click_and_drop_api.models import (
    AddressRequest,
    AddressReturns,
    BillingDetailsRequest,
    CreateOrderRequest,
    CreateOrdersResponse,
    CustomerReference,
    GetOrderInfoResource,
    PostageDetailsRequest,
    RecipientDetailsRequest,
    ReturnShipment as RMReturnShipment,
    ReturnsRequest,
    Service,
    ShipmentPackageRequest,
)
from royal_mail_combined.click_and_drop_api.models.return_models import ReturnRequestContainer, ReturnResponseContainer
from royal_mail_combined.converters_no_import import tracking_link
from royal_mail_combined.core.consts_types import PackageFormat, RoyalMailServiceCodes, SendNotifcationsTo

from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.responses import CompletedShipmentResponse
from shipaw.models.address import Address, Contact, FullContact
from shipaw.utils.consts_enums import ShipDirection
from shipaw.models.shipment import Shipment
from shipaw.utils.funcs import date_to_datetime


def build_booking_response_inbound(rm_response: ReturnResponseContainer, shipment: Shipment, label_data: bytes):
    return CompletedShipmentResponse(
        label_data=label_data,
        shipment=shipment,
        shipment_num=rm_response.unique_ids_str,
        shipment_numbers=rm_response.unique_ids,
        tracking_links=rm_response.tracking_links,
        collection_id=rm_response.collection_response.collection_order_id,
        data=rm_response.model_dump(),
        status='Success',
        success=True,
    )


def print_response_errors(rm_response: CreateOrdersResponse):
    logger.error(f'Errors booking outbound shipment: {pformat(rm_response.failed_orders, indent=2, width=120)}')


def build_booking_response_outbound_f_fetched(
    fetched: list[GetOrderInfoResource], shipment: Shipment, label_data: bytes
) -> CompletedShipmentResponse:
    tracking_numbers = [order.tracking_number for order in fetched]
    tracking_links = [tracking_link(_) for _ in tracking_numbers]
    idents = [_.order_identifier for _ in fetched]
    idents_s = map(str, idents)
    res = CompletedShipmentResponse(
        order_identifiers=idents,
        shipment=shipment,
        label_data=label_data,
        shipment_num=';'.join(idents_s),
        shipment_numbers=tracking_numbers,
        tracking_links=tracking_links,
        data={_.order_identifier: _.model_dump() for _ in fetched},
        status='Success',
        success=True,
    )
    return res


def build_booking_response_outbound(
    rm_response: CreateOrdersResponse, shipment: Shipment, label_data: bytes
) -> CompletedShipmentResponse:
    success = rm_response.errors_count == 0
    tracking_numbers = [order.tracking_number for order in rm_response.created_orders]
    tracking_links = [tracking_link(_) for _ in tracking_numbers]
    res = CompletedShipmentResponse(
        shipment=shipment,
        label_data=label_data,
        shipment_num=rm_response.success_idents_str,
        shipment_numbers=rm_response.success_ident_strs,
        tracking_links=tracking_links,
        data={_.order_identifier: _.model_dump() for _ in rm_response.created_orders},
        status='Success' if success else 'FAIL',
        success=success,
    )
    return res


def outbound_shipment_from_agnostic(shipment: Shipment, service_code: RoyalMailServiceCodes) -> CreateOrderRequest:
    billing_details = rm_billing_details_from_fc(SHIPAW_SETTINGS.full_contact)
    postage_details = create_postage_details(shipment=shipment, service_code=service_code)

    return CreateOrderRequest(
        order_reference=shipment.shipping_date.strftime('%d/%m') + ' ' + shipment.reference[0:34],
        postage_details=postage_details,
        billing=billing_details,
        recipient=rm_recipient_details_from_agnostic_fc(shipment.recipient),
        order_date=date_to_datetime(shipment.shipping_date),
        planned_despatch_date=date_to_datetime(shipment.shipping_date),
        subtotal=0,
        shipping_cost_charged=0,
        total=0,
        packages=create_packages(
            num_parcels=shipment.boxes, package_format=shipment.package_format, weight_kg=shipment.weight_kg
        ),
    )


def inbound_shipment_from_agnostic(shipment: Shipment, service_code: RoyalMailServiceCodes) -> ReturnRequestContainer:
    reqs = [
        ReturnsRequest(
            service=Service(service_code=service_code),
            shipment=RMReturnShipment(
                recipient_address=returns_address_from_agnostic_fc(shipment.recipient),
                sender_address=returns_address_from_agnostic_fc(shipment.sender),
                customer_reference=CustomerReference(reference=shipment.reference),
            ),
        )
        for _ in range(shipment.boxes)
    ]
    return ReturnRequestContainer(return_requests=reqs)


def returns_address_from_agnostic_fc(full_contact: FullContact):
    names = full_contact.contact.contact_name.split()
    first = names[0]
    last = ' '.join(names[1:]) if len(names) > 1 else ''
    return AddressReturns(
        title='',
        first_name=first,
        last_name=last,
        company_name=full_contact.address.business_name,
        address_line1=full_contact.address.address_lines[0],
        address_line2=full_contact.address.address_lines[1] if len(full_contact.address.address_lines) > 1 else None,
        address_line3=full_contact.address.address_lines[2] if len(full_contact.address.address_lines) > 2 else None,
        city=full_contact.address.town,
        county=full_contact.address.county,
        postcode=full_contact.address.postcode,
        country=full_contact.address.country,
        email=full_contact.contact.email_address,
    )


def create_postage_details(shipment: Shipment, service_code):
    send_to = (
        SendNotifcationsTo.RECIPIENT if shipment.direction == ShipDirection.OUTBOUND else SendNotifcationsTo.BILLING
    )
    return PostageDetailsRequest(
        service_code=service_code,
        send_notifications_to=send_to,
        receive_email_notification=True,
        receive_sms_notification=True,
    )


def create_packages(*, num_parcels: int, package_format: PackageFormat, weight_kg: int):
    return [
        ShipmentPackageRequest(
            weight_in_grams=weight_kg * 1000,
            package_format_identifier=package_format,
        )
        for _ in range(num_parcels)
    ]


def rm_address_from_agnostic_fc(full_contact: FullContact) -> AddressRequest:
    return AddressRequest(
        full_name=full_contact.contact.contact_name,
        company_name=full_contact.address.business_name,
        address_line1=full_contact.address.address_lines[0],
        address_line2=full_contact.address.address_lines[1] if len(full_contact.address.address_lines) > 1 else None,
        address_line3=full_contact.address.address_lines[2] if len(full_contact.address.address_lines) > 2 else None,
        city=full_contact.address.town,
        county=full_contact.address.county,
        postcode=full_contact.address.postcode,
        country_code=full_contact.address.country,
    )


def rm_recipient_details_from_agnostic_fc(full_contact: FullContact) -> RecipientDetailsRequest:
    return RecipientDetailsRequest(
        address=rm_address_from_agnostic_fc(full_contact),
        phone_number=full_contact.contact.mobile_phone or full_contact.contact.phone_number,
        email_address=full_contact.contact.email_address,
    )


def rm_billing_details_from_fc(full_contact: FullContact):
    return BillingDetailsRequest(
        address=rm_address_from_agnostic_fc(full_contact),
        phone_number=full_contact.contact.phone_number,
        email_address=full_contact.contact.email_address,
    )


def full_contact_from_rm(recipient: RecipientDetailsRequest) -> FullContact:
    return FullContact(
        contact=Contact(
            contact_name=recipient.address.full_name,
            phone_number=recipient.phone_number,
            email_address=recipient.email_address,
            mobile_phone=recipient.phone_number,
        ),
        address=Address(
            business_name=recipient.address.company_name,
            address_lines=[
                line
                for line in [
                    recipient.address.address_line1,
                    recipient.address.address_line2,
                    recipient.address.address_line3,
                ]
                if line
            ],
            town=recipient.address.city,
            county=recipient.address.county,
            postcode=recipient.address.postcode,
            country=recipient.address.country_code,
        ),
    )



from __future__ import annotations

from base64 import b64encode
from pathlib import Path

from pydantic import ConfigDict, Field, computed_field

from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.alerts import Alerts
from shipaw.models.base import ShipawBaseModel
from shipaw.utils.label_file import get_label_stem, unused_path
from shipaw.models.shipment import Shipment


class ShipawTemplate(ShipawBaseModel):
    template_path: str
    context: dict = Field(default_factory=dict)

    def render_template(self, request):
        if not self.template_path:
            raise ValueError('No template_path set')
        return request.app.shipaw_settings.templates.TemplateResponse(
            request=request, name=self.template_path, context=self.context
        )


class BaseResponse(ShipawBaseModel):
    alerts: Alerts = Alerts.empty()
    data: dict | None = None
    success: bool | None = None
    status: str | None = None
    template: ShipawTemplate | None = None


class ShipmentResponse(BaseResponse):
    shipment: Shipment
    label_data: bytes | None = None
    _label_path: Path | None = None
    shipment_num: str | None = None
    shipment_numbers: list[str] = Field(default_factory=list)
    order_identifiers: list[int] = Field(default_factory=list)
    tracking_links: list[str] = Field(default_factory=list)
    collection_id: str | None = None

    model_config = ConfigDict(json_encoders={bytes: lambda v: b64encode(v).decode('utf-8') if v else None})

    @computed_field
    @property
    def label_path(self) -> Path:
        if self._label_path is None:
            folder = SHIPAW_SETTINGS.label_dir / self.shipment.direction
            label_stem = get_label_stem(self.shipment)
            label_filepath = (folder / label_stem).with_suffix('.pdf')
            self._label_path = unused_path(label_filepath)
        return self._label_path


class CompletedShipmentResponse(ShipmentResponse):
    # model_config = ConfigDict(frozen=True)
    shipment: Shipment
    label_data: bytes
    shipment_num: str
    shipment_numbers: list[str]
    tracking_links: list[str]
    collection_id: str | None = None


class ShipawTemplateResponse(BaseResponse):
    template: ShipawTemplate



from typing import Any
from urllib.parse import unquote

from fastapi import APIRouter, Body, Depends
from loguru import logger
from royal_mail_combined.parcels_apis.address.models import (
    AddressRecordDef,
    AddressRecordDefPermissive,
    AddressSummaryDef,
)
from starlette.requests import Request
from starlette.responses import JSONResponse, StreamingResponse

from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.alerts import Alerts
from shipaw.fapi.backend import (
    errored_shipment,
    maybe_alert_apc,
    notify_dev,
    notify_version,
    resize_and_write_labels,
    try_book_shipment,
)
from shipaw.fapi.form_data import provider_from_form, shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import CompletedShipmentResponse, ShipawTemplate, ShipawTemplateResponse
from shipaw.fapi.ui_funcs import make_nice_str
from shipaw.logging import log_obj, log_obj_text
from shipaw.models.address import Address
from shipaw.models.shipment import Shipment
from shipaw.providers.registry import PROVIDER_REGISTER
from shipaw.utils.consts_enums import RM_UNAVAIL
from shipaw.utils.funcs import compare_texts

router = APIRouter()
NoAddressFound = AddressRecordDefPermissive(label='No matching results', address_id='')


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def shipping_form_api(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj_text(shipment, 'Shipment received at /ship_form:')

    alerts: Alerts = request.app.alerts
    alerts += notify_dev()
    alerts += await notify_version(request)

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj_text(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    # check phone number
    # alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    # check not apc + dropoff
    alerts = await maybe_alert_apc(shipment_request)
    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response: CompletedShipmentResponse = await try_book_shipment(shipment_request)
    if shipment_response.alerts.errors:
        return await errored_shipment(shipment_response)
    log_obj(shipment_response, 'Shipment Booked')

    await resize_and_write_labels(shipment_response.label_data, shipment_response.label_path)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'shipment_request': shipment_request, 'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.get('/providers', response_class=JSONResponse)
async def providers():
    dflt = SHIPAW_SETTINGS.default_provider_name
    available = sorted(PROVIDER_REGISTER.keys(), key=lambda p: p != dflt)
    provider_response = {make_nice_str(_): _ for _ in available}
    return JSONResponse(provider_response)


@router.get('/provider_directions/{provider_name}', response_class=JSONResponse)
async def provider_directions(provider_name: str):
    provider = await provider_from_form(provider_name)
    directions = provider.valid_directions
    dir_dict = {make_nice_str(_.value): _.value for _ in directions}
    return JSONResponse(dir_dict)


@router.get('/provider_direction_services/{provider_name}/{direction}', response_class=JSONResponse)
async def provider_direction_services(provider_name: str, direction: str):
    provider = await provider_from_form(provider_name)
    dflt = provider.default_service
    available = sorted(provider.valid_direction_services.get(direction, []), key=lambda s: s.value != dflt)
    res_dir = {make_nice_str(_.name): _.value for _ in available}
    logger.debug(f'Provider Services: {res_dir}')
    return JSONResponse(res_dir)


@router.get('/provider_direction_formats/{provider_name}/{direction}', response_class=JSONResponse)
async def provider_direction_formats(provider_name: str, direction: str):
    provider = await provider_from_form(provider_name)
    available = provider.valid_direction_formats.get(direction, [None])
    res_dir = {make_nice_str(_.name): _.value for _ in available}
    logger.debug(f'Provider Formats: {res_dir}')
    return JSONResponse(res_dir)


@router.get('/address_search/{search_text}', response_model=list[AddressSummaryDef])
async def address_search(search_text: str):
    provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
    res = provider.client.address_search(search_text)
    addresses = [_ for _ in res.addresses if _.type == 'Address']
    other = [_ for _ in res.addresses if _.type != 'Address']
    logger.debug(f'Search for "{search_text}" returned {len(addresses)} addresses and {len(other)} other results')
    return res.addresses


def address_search_text(address: Address) -> str:
    fields = [address.business_name] + address.address_lines + [address.town, address.postcode]
    return ', '.join([_ for _ in fields if _])


def match_addr_type(addr: AddressSummaryDef, expected_type='Address') -> bool:
    if addr.type != expected_type:
        logger.warning(f'Skipping "{addr.type}" type: {addr.summary}')
        return False
    return True


@router.post('/address_search_full', response_model=list[AddressRecordDefPermissive])
async def address_search_full(address: Address):
    search_text = address_search_text(address)
    if hits := await get_hits(address.postcode, search_text):
        return hits
    address.business_name = ''
    if hits := await get_hits(address.postcode, address_search_text(address)):
        return hits
    else:
        return [NoAddressFound]


async def get_hits(postcode: str, search_text: str) -> list[Any]:
    provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
    if not provider:
        logger.info(RM_UNAVAIL)
        return [AddressRecordDef(label=RM_UNAVAIL, address_id='')]
    addresses = await address_search(search_text)
    hits = []
    for addr in addresses:
        if match_addr_type(addr, 'Address'):
            retrieved: AddressRecordDef = provider.client.address_retrieve(addr.address_id)
            if compare_texts(retrieved.postal_code, postcode):
                hits.append(retrieved)
    await log_address_hits(postcode, hits)
    return hits


async def log_address_hits(postcode: str, hits: list[Any]):
    logger.debug(
        f'{len(hits)} Address{"es" if len(hits) != 1 else ""} matched postcode "{postcode}"'
        f' {"\n\t".join([addr.label.replace("\n", ",") for addr in hits])}'
    )


@router.get('/address_retrieve/{addr_id}', response_model=AddressRecordDefPermissive)
async def address_retrieve(addr_id: str):
    provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
    if not provider:
        logger.info(RM_UNAVAIL)
        return AddressRecordDef(label=RM_UNAVAIL, address_id='')
    addr_id = unquote(addr_id)  # todo is this required?
    res: AddressRecordDef = provider.client.address_retrieve(addr_id)
    logger.debug(f'Address Retrieval for "{addr_id}" returned: {res.label.replace("\n", ",")}')
    return res


@router.get('/logs/stream')
async def logs_stream(request: Request):
    stream = request.app.state.log_stream

    async def event_source():
        async for chunk in stream.stream(replay=100):
            if await request.is_disconnected():
                break
            yield chunk

    return StreamingResponse(
        event_source(),
        media_type='text/event-stream',
        headers={
            'Cache-Control': 'no-cache',
            'Connection': 'keep-alive',
            'X-Accel-Buffering': 'no',
        },
    )


# @router.get('/address_search_pc', response_model=list[AddressRecordDefPermissive])
# async def address_search_pc(postcode: str, search_text: str):
#     postcode = postcode.strip()
#     provider = PROVIDER_REGISTER.get('ROYAL_MAIL')
#     if not provider:
#         logger.info(RM_UNAVAIL)
#         return [AddressRecordDef(label=RM_UNAVAIL, address_id='')]
#     addresses = await address_search(f'{search_text}, {postcode}')
#     hits = []
#     for addr in addresses:
#         if not addr.type == 'Address':
#             logger.warning(f'Skipping "{addr.type}" type: {addr.summary}')
#             continue
#         retrieved: AddressRecordDef = provider.client.address_retrieve(addr.address_id)
#         if compare_texts(retrieved.postal_code, postcode):
#             hits.append(retrieved)
#     logger.debug(
#         f'{len(hits)} Address{"es" if len(hits) != 1 else ""} matched postcode "{postcode}":\n'
#         f' {"\n".join([addr.label.replace("\n", ",") for addr in hits])}'
#     )
#     if len(hits) == 0:
#         hits = [AddressRecordDefPermissive(label='No matching results', address_id='')]
#     return hits



import time
from collections.abc import Callable
from datetime import date, datetime
from typing import Any


def wait_for(
    func: Callable, *args: Any, tries: int = 10, wait_for_type: type = Any, is_not_none: bool = True, **kwargs
) -> Any:
    for _ in range(tries):
        time.sleep(1)
        try:
            result = func(*args, **kwargs)
            if is_not_none:
                assert result is not None, 'Data not ready, retrying...'
            if wait_for_type is not Any:
                assert isinstance(result, wait_for_type), f'Expected type {wait_for_type}, got {type(result)}'
            return result
        except AssertionError:
            continue
    raise RuntimeError(f'Data not ready after {tries} retries')


# def wait_for[T: Any](func: Callable[Any, T], *args, tries=10) -> T:
#     for i in range(tries):
#         try:
#             return_data = func(*args)
#             assert return_data is not None, 'Data not ready , retrying...'
#             return return_data
#         except AssertionError:
#             pass
#     raise RuntimeError(f'Data not ready after {tries} retries')
def strip_text(text: str):
    return text.replace(' ', '').lower()


def compare_texts(text1: str, text2: str, ignore_case: bool = True, ignore_whitespace: bool = True) -> bool:
    text1 = text1.lower() if ignore_case else text1
    text2 = text2.lower() if ignore_case else text2
    text1 = text1.replace(' ', '') if ignore_whitespace else text1
    text2 = text2.replace(' ', '') if ignore_whitespace else text2
    return text1 == text2


def date_to_datetime(d: date) -> datetime:
    return datetime.combine(d, datetime.min.time())



import json
import pprint
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger
from pydantic import BaseModel

from shipaw.config import SHIPAW_SETTINGS

if TYPE_CHECKING:
    pass


def ndlog_dict(data: dict, ndjson_file: Path | None = None):
    ndjson_file = ndjson_file or SHIPAW_SETTINGS.ndjson_log_file
    with open(ndjson_file, 'a') as jf:
        print(json.dumps(data, separators=(',', ':')), file=jf)


def log_obj_text(obj: BaseModel, message: str = '', *, level: str = 'DEBUG', logger_=logger):
    message = message or obj.__class__.__name__
    model_data = obj.model_dump(mode='json', exclude={'label_data': ..., 'response': {'label_data'}})
    msg = f'{message}:\n{pprint.pformat(model_data, indent=2)}'.replace('{', r'{{').replace('}', r'}}')
    logger_.opt(depth=2).log(
        level,
        msg,
    )


def log_obj_json(obj: BaseModel, message: str = '', *, ndjson_file=None):
    ndjson_file = ndjson_file or SHIPAW_SETTINGS.ndjson_log_file
    timestamp = datetime.now().isoformat(timespec='seconds')
    logdict = {
        'data_type': type(obj).__name__,
        'timestamp': timestamp,
        'message': message,
        'obj_data': obj.model_dump(mode='json', exclude={'label_data': ..., 'response': {'label_data'}}),
    }
    ndlog_dict(logdict, ndjson_file=ndjson_file)


def log_obj(
    obj: BaseModel,
    message: str = None,
    level: str = 'DEBUG',
    logger_=logger,
    ndjson_file=None,
):
    log_obj_text(obj, message, level=level, logger_=logger_)
    log_obj_json(obj, message, ndjson_file=ndjson_file)



from abc import ABC, abstractmethod
from enum import StrEnum
from typing import ClassVar, TYPE_CHECKING

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings

from shipaw.models.base import ShipawBaseModel
from shipaw.utils.consts_enums import PackageFormat, ShipDirection
from shipaw.models.shipment import Shipment

if TYPE_CHECKING:
    from shipaw.fapi.responses import CompletedShipmentResponse
    from shipaw.fapi.requests import ShipmentRequest


class ProviderName(StrEnum):
    ROYAL_MAIL = 'ROYAL_MAIL'
    APC = 'APC'


class HasServiceCodes(ABC):
    service_codes_type: ClassVar[type[StrEnum]]
    default_service: ClassVar[StrEnum]

    @classmethod
    def lookup_service(cls, service_name: str):
        return cls.service_codes_type[service_name]

    @classmethod
    def reverse_lookup_service(cls, service_code: str):
        return cls.service_codes_type(service_code).name

    @classmethod
    def services_as_tups(cls) -> list[tuple[str, str]]:
        return [(e.name, str(e.value)) for e in cls.service_codes_type]

    @classmethod
    def services_as_dict(cls) -> dict[str, str]:
        return dict(cls.service_codes_type.__members__.items())


class ShippingProvider(HasServiceCodes, ABC, ShipawBaseModel):
    name: ClassVar[ProviderName]
    settings: BaseSettings
    settings_type: ClassVar[type[BaseSettings]]
    valid_directions: ClassVar[list[ShipDirection]]
    valid_direction_services: ClassVar[dict[ShipDirection, list[StrEnum]]] = Field(default_factory=dict)
    valid_direction_formats: ClassVar[dict[ShipDirection, list[PackageFormat]]] = Field(default_factory=dict)

    @property
    @abstractmethod
    def client(self):
        raise NotImplementedError

    @abstractmethod
    def is_sandbox(self) -> bool: ...

    @staticmethod
    @abstractmethod
    def provider_shipment(shipment: Shipment, service_code: StrEnum) -> BaseModel:
        """Takes Shipaw's agnostic Shipment object and returns provider-specific Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def agnostic_shipment(shipment: BaseModel) -> Shipment:
        """Takes provider-specific Shipment object and returns Shipaw's agnostic Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def book_shipment_request(shipment_request: 'ShipmentRequest') -> 'CompletedShipmentResponse': ...



import datetime as dt

from pydantic import Field

from shipaw.models.address import FullContact
from shipaw.models.base import ShipawBaseModel
from shipaw.utils.consts_enums import PackageFormat, ShipDirection


class Shipment(ShipawBaseModel):
    recipient: FullContact
    sender: FullContact | None = None  # default to ShipawSettings.address/contact if None

    boxes: int = 1
    shipping_date: dt.date
    direction: ShipDirection
    own_label: bool | None = None  # print own label vs driver brings
    # service: AgnostServiceName = AgnostServiceName.NEXT_DAY
    reference: str = ''

    # todo is context still used?
    context: dict = Field(default_factory=dict)

    collect_ready: dt.time = dt.time(hour=9, minute=0)
    collect_closed: dt.time = dt.time(hour=17, minute=0)

    package_format: PackageFormat = PackageFormat.PARCEL
    weight_kg: int = 10

    @property
    def remote_full_contact(self) -> FullContact:
        match self.direction:
            case ShipDirection.OUTBOUND:
                return self.recipient
            case ShipDirection.INBOUND | ShipDirection.DROPOFF:
                return self.sender
            case _:
                raise ValueError('Bad ShipDirection')


def sample_shipment() -> Shipment:
    from shipaw.models.address import Address, Contact

    contact = Contact(
        contact_name='Test Contact name',
        mobile_phone='07666666666',
        email_address='sdgsdg@sdgsdg.com',
    )
    address = Address(
        postcode='DA16 3HU',
        address_lines=['25 Bennet Close'],
        town='Welling',
        country='GB',
        business_name='Test Company',
    )
    full_contact = FullContact(contact=contact, address=address)
    return Shipment(
        recipient=full_contact,
        boxes=2,
        shipping_date=dt.date.today() + dt.timedelta(days=2),
        direction=ShipDirection.OUTBOUND,
        reference='Test Reference',
        # service=AgnostServiceName.NEXT_DAY,
    )



from .providers import APCSettings, APCShippingProvider, RoyalMailProvider, RoyalMailSettingsGlobal

__all__ = [
    'APCSettings',
    'APCShippingProvider',
    'RoyalMailProvider',
    'RoyalMailSettingsGlobal',
]



from __future__ import annotations

from pathlib import Path

from apc_hypaship.error import apc_http_status_alerts
from httpx import HTTPStatusError
from loguru import logger
from pawdf.array_pdf.array_p import on_a4

from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import CompletedShipmentResponse, ShipawTemplate, ShipawTemplateResponse, ShipmentResponse
from shipaw.logging import log_obj
from shipaw.utils.consts_enums import ShipDirection

from shipaw.providers.provider_abc import ProviderName


async def try_book_shipment(shipment_request: ShipmentRequest) -> CompletedShipmentResponse:
    shipment_response = ShipmentResponse(shipment=shipment_request.shipment)
    try:
        shipment_response = shipment_request.provider.book_shipment_request(shipment_request)

    except HTTPStatusError as e:
        await maybe_apc_response_error(e, shipment_request, shipment_response)

    except Exception as e:
        logger.exception(f'Error booking shipment: {e}')
        shipment_response.alerts += Alert.from_exception(e)

    return shipment_response


# async def try_get_write_label(request: ShipmentRequest, response: ShipmentResponse):
#     if not response.label_data:
#         await try_get_label_data(request, response)
#
#     try:
#         await resize_and_write_labels(response.label_data, response.label_path)
#
#     except Exception as e:
#         logger.exception(f'Error writing label file: {e}')
#         response.alerts += Alert.from_exception(e)


# async def try_get_label_data(request: ShipmentRequest, response: ShipmentResponse) -> None:
#     if response.label_data is not None:
#         logger.info('Label data already present, not fetching')
#         return
#     try:
#         if response.shipment_num:
#             response.label_data = await request.provider.wait_fetch_label_async(response.shipment_num)
#         else:
#             logger.warning('No shipment number to fetch label data')
#
#     except HTTPStatusError as e:
#         await maybe_apc_response_error(e, request, response)
#
#     except Exception as e:
#         logger.exception('Error getting label data')
#         response.alerts += Alert.from_exception(e)


async def maybe_apc_response_error(e: HTTPStatusError, shipment_request, shipment_response):
    if shipment_request.provider_name == ProviderName.APC:
        for alert in await apc_http_status_alerts(e):
            shipment_response.alerts += alert
    else:
        logger.exception(e)
        shipment_response.alerts += Alert.from_exception(e)


async def maybe_alert_apc(shipment_request):
    alerts = Alerts.empty()
    if (
        shipment_request.provider_name == ProviderName.APC
        and shipment_request.shipment.direction == ShipDirection.DROPOFF
    ):
        alerts += Alert(
            message='APC does not support drop-off shipments - please select Outbound or Inbound Collection',
            type=AlertType.ERROR,
        )
    return alerts


async def resize_and_write_labels(label_content: bytes, label_path: Path):
    og_size_path = label_path.parent / 'original_size' / label_path.name
    og_size_path.parent.mkdir(parents=True, exist_ok=True)
    og_size_path.write_bytes(label_content)
    logger.info(f'Resizing {og_size_path} to A4 at {label_path}')
    on_a4(input_file=og_size_path, output_file=label_path)
    logger.info(f'Wrote label to {label_path}')


def get_version():
    from importlib.metadata import version, PackageNotFoundError

    try:
        return version('shipaw')
    except PackageNotFoundError:
        return 'unknown'


async def notify_version(request):
    alerts = Alerts.empty()
    sandbox = request.app.shipaw_settings.shipper_live is False
    if sandbox:
        live_msg = 'Test Mode - No Shipments will be booked'
        notification_type = AlertType.NOTIFICATION
    else:
        live_msg = 'Live Mode - Real Shipments will be booked'
        notification_type = AlertType.WARNING

    msg = f'Shipaw Version {get_version()} is in {live_msg}'
    logger.info(msg)
    alerts += Alert(message=msg, type=notification_type)
    return alerts


def notify_dev() -> Alerts:
    alerts = Alerts.empty()
    if any(['prdev' in str(_).lower() for _ in Path(__file__).parents]):
        msg = '"prdev" in cwd tree - BETA MODE - This is a development version'
        logger.info(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)
    return alerts


async def errored_shipment(shipment_response):
    log_obj(shipment_response.alerts, 'Errors booking shipment:')
    alerts = shipment_response.alerts
    shipment_response.template = ShipawTemplate(
        template_path='/alerts.html',
        context={'alerts': alerts},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)



from __future__ import annotations

import re
import datetime as dt
from enum import StrEnum
from typing import Annotated

import phonenumbers
from loguru import logger
from pydantic import AfterValidator, BeforeValidator, Field, ValidationError


#
# class ShipDirection(StrEnum):
#     INBOUND = 'in'
#     OUTBOUND = 'out'
#     DROPOFF = 'dropoff'

TOD = dt.date.today()
COLLECTION_CUTOFF = dt.time(23, 59, 59)
ADVANCE_BOOKING_DAYS = 28
WEEKDAYS_IN_RANGE = [
    TOD + dt.timedelta(days=i) for i in range(ADVANCE_BOOKING_DAYS) if (TOD + dt.timedelta(days=i)).weekday() < 5
]

COLLECTION_WEEKDAYS = [i for i in WEEKDAYS_IN_RANGE if not i == TOD]


def limit_daterange_no_weekends(v: dt.date) -> dt.date:
    logger.debug(f'Validating date: {v}')
    if v:
        if isinstance(v, str):
            logger.debug(f'parsing date string assuming isoformat: {v}')
            v = dt.date.fromisoformat(v)

        if isinstance(v, dt.date):
            if v < TOD or v.weekday() > 4:
                logger.debug(f'Date {v} is a weekend or in the past - using next weekday')
                v = min(WEEKDAYS_IN_RANGE)

            if v > max(WEEKDAYS_IN_RANGE):
                logger.debug(f'Date {v} is too far in the future - using latest weekday (max 28 days in advance)')
                v = max(WEEKDAYS_IN_RANGE)

    return v


def validate_phone(v: str, values) -> str:
    logger.warning(f'Validating phone: {v}')
    phone = v.replace(' ', '')
    nummy = phonenumbers.parse(phone, 'GB')
    assert phonenumbers.is_valid_number(nummy)
    return phonenumbers.format_number(nummy, phonenumbers.PhoneNumberFormat.E164)


POSTCODE_PATTERN = re.compile(r'([A-Z]{1,2}\d[A-Z\d]? ?\d[A-Z]{2})')

pc_excluded = {'C', 'I', 'K', 'M', 'O', 'V'}


def validate_uk_postcode(v: str):
    if not re.match(POSTCODE_PATTERN, v) and not set(v[-2:]).intersection(pc_excluded):
        raise ValidationError('Invalid UK postcode')
    return v


VALID_POSTCODE = Annotated[
    str,
    AfterValidator(validate_uk_postcode),
    BeforeValidator(lambda s: s.strip().upper()),
    Field(..., description='A valid UK postcode'),
]


class ShipDirection(StrEnum):
    INBOUND = 'Inbound'
    OUTBOUND = 'Outbound'
    DROPOFF = 'Dropoff'


class PackageFormat(StrEnum):
    SMALL_PARCEL = 'smallParcel'
    MEDIUM_PARCEL = 'mediumParcel'
    PARCEL = 'parcel'
    LETTER = 'letter'
    LARGE_LETTER = 'largeLetter'
    DOCUMENTS = 'documents'
    UNDEFINED = 'undefined'


def enum_as_tups(enum_type: type[StrEnum]) -> list[tuple[str, str]]:
    return list(enum_type.__members__.items())


def enum_as_dict(enum_type: type[StrEnum]) -> dict[str, str]:
    return dict(enum_type.__members__.items())


def enum_lookup(*, enum_type: type[StrEnum], attr_name: str) -> str:
    res = getattr(enum_type, attr_name, None)
    if not res:
        raise ValueError(f'Invalid attr_name: {attr_name}')
    return res


def enum_reverse_lookup(*, enum_type: type[StrEnum], attr_value: str) -> str:
    res = next(name for name in enum_type.__dict__.values() if name == attr_value)
    if not res:
        raise ValueError(f'Invalid service code: {attr_value}')
    return res


RM_UNAVAIL = 'Royal Mail Unavailable = No address searches.'



from __future__ import annotations

import json
from datetime import date, time

from fastapi import Depends, Form, HTTPException
from loguru import logger

# from pawdantic.paw_types import VALID_POSTCODE
from pydantic import EmailStr

from shipaw.models.address import Address, Contact, FullContact
from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.requests import ShipmentRequest
from shipaw.utils.consts_enums import PackageFormat, ShipDirection, VALID_POSTCODE
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ProviderName, ShippingProvider
from shipaw.providers.registry import PROVIDER_REGISTER


async def full_contact_form(
    address_line1: str = Form(...),
    address_line2: str = Form(''),
    address_line3: str = Form(''),
    town: str = Form(...),
    postcode: VALID_POSTCODE = Form(...),
    contact_name: str = Form(...),
    email_address: EmailStr = Form(...),
    business_name: str = Form(...),
    mobile_phone: str = Form(...),
) -> FullContact:
    return FullContact(
        address=Address(
            address_lines=[address_line1, address_line2, address_line3],
            town=town,
            postcode=postcode,
            business_name=business_name,
        ),
        contact=Contact(
            contact_name=contact_name,
            email_address=email_address,
            mobile_phone=mobile_phone.strip().replace(' ', ''),
        ),
    )


async def shipment_from_form(
    full_contact: FullContact = Depends(full_contact_form),
    shipping_date: date = Form(...),
    boxes: int = Form(...),
    direction: ShipDirection = Form(...),
    reference: str = Form(...),
    context_json: str = Form(...),
    collect_ready: int = Form(9),
    collect_closed: int = Form(17),
    own_label: bool = Form(True),
    package_format: PackageFormat = Form(PackageFormat.PARCEL),
    weight_kg: int = Form(10),
) -> Shipment:
    collect_ready = time(hour=collect_ready)
    collect_closed = time(hour=collect_closed)
    context = json.loads(context_json)
    logger.debug('Creating Shipment Request from form')

    if direction == ShipDirection.OUTBOUND:
        sender = None
        recipient = full_contact
        own_label = None
    elif direction in {ShipDirection.INBOUND, ShipDirection.DROPOFF}:
        sender = full_contact
        recipient = SHIPAW_SETTINGS.full_contact
        own_label = own_label
    else:
        raise ValueError(f'Unknown direction: {direction}')

    shipment = Shipment(
        recipient=recipient,
        sender=sender,
        boxes=boxes,
        shipping_date=shipping_date,
        direction=direction,
        reference=reference,
        context=context,
        collect_ready=collect_ready,
        collect_closed=collect_closed,
        own_label=own_label,
        package_format=package_format,
        weight_kg=weight_kg,
    )
    return shipment


async def provider_from_form(provider_name: str = Form(...)):
    provider = PROVIDER_REGISTER.get(provider_name)
    if not provider:
        raise HTTPException(status_code=404, detail='Provider not found')
    return provider


async def shipment_request_form(
    provider: ShippingProvider = Depends(provider_from_form),
    shipment: Shipment = Depends(shipment_from_form),
    provider_name: ProviderName = Form(...),
    service: str = Form(...),
) -> ShipmentRequest:
    serv = provider.service_codes_type(service)
    return ShipmentRequest(shipment=shipment, provider_name=provider_name, service_code=serv)


async def shipment_form_json(shipment_json: str = Form(...)) -> Shipment:
    shipy = Shipment.model_validate_json(shipment_json)
    return shipy


async def shipment_request_form_json(shipment_request_json: str = Form(...)) -> ShipmentRequest:
    shipy = ShipmentRequest.model_validate_json(shipment_request_json)
    return shipy


async def context_form_json(context_json: str = Form(...)) -> dict:
    context = json.loads(context_json)
    return context



import base64
from typing import ClassVar, override

from royal_mail_combined import RoyalMailClient
from royal_mail_combined.click_and_drop_api.models import (
    CreateOrderRequest,
    CreateOrdersRequest,
)
from royal_mail_combined.click_and_drop_api.models.return_models import ReturnRequestContainer
from royal_mail_combined.config import RoyalMailSettingsGlobal
from royal_mail_combined.core.consts_types import RoyalMailServiceCodes, RoyalMailInboundServiceCodes

from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import CompletedShipmentResponse
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ProviderName, ShippingProvider
from shipaw.providers.registry import register_provider_type
from shipaw.providers.royal_mail.royal_mail_funcs import (
    build_booking_response_inbound,
    build_booking_response_outbound_f_fetched,
    full_contact_from_rm,
    inbound_shipment_from_agnostic,
    outbound_shipment_from_agnostic,
    print_response_errors,
)
from shipaw.utils.consts_enums import PackageFormat, ShipDirection
from shipaw.utils.label_file import merge_pdf_bytes


@register_provider_type
class RoyalMailProvider(ShippingProvider):
    settings: RoyalMailSettingsGlobal

    name: ClassVar[ProviderName] = ProviderName.ROYAL_MAIL
    settings_type: ClassVar[type[RoyalMailSettingsGlobal]] = RoyalMailSettingsGlobal
    service_codes_type: ClassVar[type[RoyalMailServiceCodes]] = RoyalMailServiceCodes
    default_service: ClassVar[RoyalMailServiceCodes] = RoyalMailServiceCodes.EXPRESS_24
    valid_directions: ClassVar[list[ShipDirection]] = [
        ShipDirection.OUTBOUND,
        ShipDirection.INBOUND,
        ShipDirection.DROPOFF,
    ]

    valid_direction_services = {
        ShipDirection.OUTBOUND: [
            RoyalMailServiceCodes.TRACKED_24,
            RoyalMailServiceCodes.EXPRESS_24,
            RoyalMailServiceCodes.SPECIAL_1PM,
            RoyalMailServiceCodes.EXPRESS_10,
            RoyalMailServiceCodes.EXPRESS_AM,
        ],
        ShipDirection.INBOUND: [RoyalMailServiceCodes.TRACKED_24_RTN],
        ShipDirection.DROPOFF: [RoyalMailServiceCodes.TRACKED_24_RTN],
    }

    valid_direction_formats = {
        ShipDirection.OUTBOUND: [PackageFormat.PARCEL, PackageFormat.SMALL_PARCEL],
        ShipDirection.INBOUND: [PackageFormat.PARCEL],
        ShipDirection.DROPOFF: [PackageFormat.PARCEL],
    }

    _client: RoyalMailClient | None = None

    @override
    def is_sandbox(self) -> bool:
        # Royal mail do not have a test environment. that's fun
        return False

    @property
    @override
    def client(self) -> RoyalMailClient:
        if self._client is None:
            if self.settings is None:
                raise ValueError('Settings must be set before using the client')
            self._client = RoyalMailClient(settings=self.settings)
        return self._client

    @classmethod
    @override
    def agnostic_shipment(cls, orders_req: CreateOrdersRequest) -> Shipment:
        order = orders_req.items[0]
        return Shipment(
            recipient=full_contact_from_rm(order.recipient),
            boxes=len(order.packages),
            shipping_date=order.planned_despatch_date.date(),
            direction=ShipDirection.OUTBOUND,
            reference=order.order_reference,
        )

    @override
    def provider_shipment(
        self, shipment: Shipment, service_code: RoyalMailServiceCodes
    ) -> CreateOrderRequest | ReturnRequestContainer:
        provider_service = self.service_codes_type(service_code)
        match shipment.direction:
            case ShipDirection.OUTBOUND:
                return outbound_shipment_from_agnostic(shipment, provider_service)
            case ShipDirection.DROPOFF | ShipDirection.INBOUND:
                return inbound_shipment_from_agnostic(shipment, provider_service)

    @override
    def book_shipment_request(self, shipment_request: ShipmentRequest) -> CompletedShipmentResponse:
        shipment = shipment_request.shipment
        service = self.service_codes_type(shipment_request.service_code)
        shipdir = shipment_request.shipment.direction
        if shipdir == ShipDirection.OUTBOUND:
            return self._book_outbound(service, shipment)
        elif shipdir in [ShipDirection.INBOUND, ShipDirection.DROPOFF]:
            return self._book_inbound_or_dropoff(service, shipment)
        else:
            raise ValueError(f'Invalid shipment direction "{shipdir}"')

    def _book_inbound_or_dropoff(self, service: RoyalMailServiceCodes, shipment: Shipment) -> CompletedShipmentResponse:
        returns_container = inbound_shipment_from_agnostic(shipment, service)
        if shipment.direction == ShipDirection.INBOUND:
            resp = self.client.book_inbound_collection(returns_container, collection_date=shipment.shipping_date)
        else:
            resp = self.client.book_inbound_dropoff(returns_container)
        labels_bytes = [base64.b64decode(order.label) for order in resp.created_orders]
        combined_label_bytes = merge_pdf_bytes(labels_bytes)
        return build_booking_response_inbound(resp, shipment, combined_label_bytes)

    def _book_outbound(self, service: RoyalMailServiceCodes, shipment: Shipment) -> CompletedShipmentResponse:
        order_create = outbound_shipment_from_agnostic(shipment, service)
        booking_response = self.client.book_outbound(order_create)
        if booking_response.errors_count > 0:
            print_response_errors(booking_response)
        label_data: bytearray = self.client.get_label_data(booking_response.success_idents_str)
        fetched_response = self.client.fetch_specific_orders(booking_response.success_idents_str)
        return build_booking_response_outbound_f_fetched(fetched_response, shipment, label_data)
        # return build_booking_response_outbound(resp, shipment, label_data)



from __future__ import annotations

import asyncio
import html
from collections import deque
from typing import AsyncIterator, TYPE_CHECKING, Any

LogLineTup = tuple[str, str]

if TYPE_CHECKING:
    from loguru import Message
else:
    Message = Any


def _sse_event(event: str, data: str) -> str:
    # SSE requires each payload line to be prefixed with "data: "
    text = data.replace('\r\n', '\n').replace('\r', '\n')
    data_lines = text.split('\n')
    payload = '\n'.join(f'data: {line}' for line in data_lines)
    return f'event: {event}\n{payload}\n\n'


class LogStream:
    def __init__(self, max_history: int = 300, queue_size: int = 200) -> None:
        self._history: deque[LogLineTup] = deque(maxlen=max_history)
        self._subscribers: set[asyncio.Queue[LogLineTup]] = set()
        self._queue_size = queue_size

    def sink(self, message: Message) -> None:
        rec = message.record
        level_name = rec['level'].name.lower()
        line = f'{rec["time"]:%H:%M:%S} | {rec["level"].name:<8} | {rec["message"]}'
        self._history.append((level_name, line))
        dead: list[asyncio.Queue[LogLineTup]] = []
        for q in self._subscribers:
            try:
                q.put_nowait((level_name, line))
            except asyncio.QueueFull:
                # Drop if a client is too slow; keep app logging non-blocking.
                pass
            except RuntimeError:
                dead.append(q)

        for q in dead:
            self._subscribers.discard(q)

    async def stream(self, replay: int = 80) -> AsyncIterator[str]:
        q: asyncio.Queue[LogLineTup] = asyncio.Queue(maxsize=self._queue_size)
        self._subscribers.add(q)

        try:
            # Replay recent history when client connects.
            for level_name, line in list(self._history)[-replay:]:
                safe = html.escape(line)
                yield _sse_event('log', f"<div class='log-line log-{level_name}'>{safe}</div>")
                # yield f"event: log\ndata: <div class='log-line log-{level_name}'>{safe}</div>\n\n"

            while True:
                try:
                    level_name, line = await asyncio.wait_for(q.get(), timeout=15)
                    safe = html.escape(line)
                    yield _sse_event('log', f"<div class='log-line log-{level_name}'>{safe}</div>")
                    # yield f"event: log\ndata: <div class='log-line log-{level_name}'>{safe}</div>\n\n"
                except TimeoutError:
                    # Keep connection alive
                    yield ': ping\n\n'
        finally:
            self._subscribers.discard(q)
